#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <vs1000.h>

#include <player.h>
#include <mappertiny.h>
#include <minifat.h>
#include <vsNand.h>
#include <audio.h>

#include <dev1000.h>

#include "mlc.h"

/*
  Only large-page FLASH supported.
*/

extern struct FsPhysical *ph;
extern struct FsMapper *map;

//extern struct Codec *cod;
//extern struct CodecServices cs;


u_int32 GetPhysical(u_int32 logical) {
    register u_int16 bl = logical / ph->eraseBlockSize;
    register int i = 0;
    while (i < mlc.bads) { /* 15 cycles per loop */
	if (mlc.bad[i] > bl)
	    break;
	bl++;
	i++;
    }
    return (logical & (ph->eraseBlockSize-1)) + (u_int32)bl * ph->eraseBlockSize;
}


struct FsMapper *FsMapMlcCreate(struct FsPhysical *physical,
				u_int16 cacheSize) {
    /* Scan erase blocks to locate the first data block,
       bad blocks, and possible replaced blocks. */
    register u_int16 eB, eraseBlockSize = physical->eraseBlockSize;
    register u_int32 sector = 0;

    memset(&mlc, 0, sizeof(mlc));
    //mlc.bads = 0;
    mlc.rootBlock =
	((physical->eraseBlockSize < 0x100) ? 0x100 : physical->eraseBlockSize)
	+ physical->eraseBlockSize /*reserve one*/;

    {/*init data buffer pointers, 11 words*/
	register int i;
	register struct REPL *r = &mlc.replace[0];
	register u_int16 *p = fmfCache;
	for (i=0;i<MAXREPL;i++) {
	    r->data = p;
	    p += 1024;
	    r++;
	}
    }

    for (eB = 0; eB < physical->eraseBlocks; eB++) {
	physical->Read(ph, (u_int32)eB * eraseBlockSize, 2, NULL, (void *)tmpBuf);
	if ((tmpBuf[0+2] & tmpBuf[8+2] & 0xff) != 0xff) {
	    mlc.bad[mlc.bads++] = eB;
	    if (mlc.bads == MAX_BADS) {
		/* too many BAD's */
		break;
	    }
	} else if ((tmpBuf[1] & 0xff) == 0x05) {
	    /* Erase block has been written to at least once. */

	    /* Datastart must be found before other data sectors. */
	    if (*(u_int32 *)&tmpBuf[4] == 0 && mlc.dataStart == 0) {
		/* logical-to-physical offset */
		mlc.dataStart = sector;
	    }
	}
	sector += eraseBlockSize;
    }
    /* leave room for mapper.v1 structure. */
    mlc.dataStart = GetPhysical(mlc.rootBlock + physical->eraseBlockSize);
    mlc.dataBlocks = (u_int32)(physical->eraseBlocks - RESERVED_AREA - mlc.bads)
	* eraseBlockSize - mlc.dataStart;
    return NULL; //&mlcMapper;
}


void main(void) {
    /* Set the leds after nand-boot! */
    PERIP(GPIO1_SET_MASK) =  LED1;      /* POWER led on */
    PERIP(GPIO1_DDR)   |=  LED1|LED2; /* SI and SO to outputs */
    PERIP(GPIO1_MODE)  &= ~(LED1|LED2); /* SI and SO to GPIO */

    /* Restore default keymap */
    currentKeyMap = sixKeyMap;

    ph = FsPhNandCreate(1);
    FsMapMlcCreate(ph, 0); /*Initialize some stuff needed by USB mode*/

    /* Check that mapper-compatible structure is valid and upto date. */
    if (((struct FsNandPhys *)ph)->nandType < 6) {
	/* Compare structure and rewrite it if necessary. */
	__y u_int32 phys = GetPhysical(mlc.rootBlock);
	u_int16 i, err = 0;
	for (i=0;i<ph->eraseBlocks;i++) {
	    u_int32 sec = GetPhysical(mlc.dataStart +
				      (u_int32)i * ph->eraseBlockSize);
	    if ((i & 1023) == 0) {
		ph->Read(ph, phys + ((i>>10)<<2), 4, mallocAreaX,
			 mallocAreaX+1024);
	    }
	    if (mallocAreaX[i & 1023] == sec / ph->eraseBlockSize) {
		/* correct */
	    } else {
		/* incorrect, update is required. */
		err++;
		break;/* ignore rest */
	    }
	}
	if (!err) {
	    /*puts("ok");*/
	} else {
	    /* Update mapping structure */
	    phys = GetPhysical(mlc.rootBlock);
	    ph->Erase(ph, phys);
	    memset(mallocAreaX, -1, 2048/2);
	    for (i=0;i<ph->eraseBlocks;i++) {
		u_int32 sec = GetPhysical(mlc.dataStart +
					  (u_int32)i * ph->eraseBlockSize);
		mallocAreaX[i & 1023] = sec / ph->eraseBlockSize;
		if ((i & 1023) == 1023 || i == ph->eraseBlocks-1) {
		    memset(mallocAreaX+1024, -1, 4*16/2);
		    mallocAreaX[1024+1] = 0xff01;
		    LargePageWrite(phys + ((i>>10)<<2),
				   mallocAreaX, mallocAreaX+1024);
		    memset(mallocAreaX, -1, 2048/2);
		}
	    }
	}
    }
    /* Then check which program to load:
       - if USB is attached, load the MLC USB mass storage code
       - otherwise load the player firmware
    */
    while (1) {
	register u_int32 start = PLAYERSTART; /* Player */
	if (USBIsAttached()) {
	    start = MLCSTART; /* USB: MLC mapper */

	    /*
	      Perform some other setup required by the MLC code
	      to save code space on mlc.c.
	     */
	    SetRate(44100U);
	    //PERIP(GPIO1_SET_MASK) =  LED2;
	    voltages[voltIoUSB] = 31; /* 3.65V */
	    voltages[voltAnaUSB] = 31; /* 3.65V */
	    PowerSetVoltages(&voltages[voltCoreUSB]);
	    /** a 2.5ms-10ms delay to let the voltages change (rise) */
	    BusyWait10();
	}
	/* Load using error correction code. Max program size 8192 bytes. */
	ph->Read(ph, start >> 9, 16/*max*/, mallocAreaX, minifatBuffer);
	BootFromX(mallocAreaX+8);
	/* should not return */
    }
    /*Using while(1) allows the compiler to drop prolog and epilog code.*/
}
