/** \file usbspk.c VS1000B/C/D USB Speaker + Mass Storage + Player demo */
// This uses interrupts to handle simultaneous USB audio and mass storage
// This version is for standard SLC nand flash (routines in ROM)

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <vs1000.h>
#include <dev1000.h>
#include <audio.h>
#include <mapper.h>
#include <physical.h>
#include <vsNand.h>
#include <mappertiny.h>
#include <minifat.h>
#include "player.h"
#include "codec.h"
#include "romfont.h"
#include "usb.h"
#include "vectors.h"
#include "scsi.h"

// Publish some VS1000 internal data structures
extern struct FsMapper *map;
extern struct FsPhysical *ph;
extern struct CodecServices cs; 
void UserInterfaceIdleHook(void); 
extern struct SCSIVARS {
  SCSIStageEnum State;
  SCSIStatusEnum Status;
  u_int16 *DataOutBuffer;
  u_int16 DataOutSize;
  u_int16 DataBuffer[32];
  u_int16 *DataInBuffer;
  u_int16 DataInSize;
  unsigned int BlocksLeftToSend;
  unsigned int BlocksLeftToReceive;
  u_int32 CurrentDiskSector;
  u_int32 mapperNextFlushed;
  u_int16 cswPacket[7];
  u_int32 DataTransferLength;
  s_int32 Residue;
  s_int16 DataDirection;
} SCSI;


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <vs1000.h>

#include <audio.h>

#include <player.h>
#include <mappertiny.h>
#include <minifat.h>
#include <vsNand.h>

#include <dev1000.h>

#include "mlc.h"


/*
  MLC 80ns 81620MB (FAT not counted)
  write: 125kB/sec  (cygwin copy) 360kB/sec (windows copy)
  read : 658kB/sec
  MLC 45ns 81620MB (FAT not counted)
  write: 125kB/sec
  read : 680kB/sec

  write: 480kB/sec
  read:  670kB/sec
 */

//#define PATCH_SCSI_COMMAND

#ifdef PATCH_SCSI_COMMAND
#include <scsi.h>
int startUnit = 1;
#endif



/*
  Only large-page FLASH supported.
*/

/*
  6kW X:0x0000..0x17ff (fmfCache).
  9kW Y:0x1c00..0x3fff (treeBuild)
  2kW X:0x2400..0x27ff (start of MallocAreaX, USBMem in X:0x2800..0x3400)
  2kW Y:0x1000..0x17ff (dcthi)
*/

#define USE_USB
#define USE_MAP

#define DEBUG_LEVEL 0

void InitMlc(void);

#if DEBUG_LEVEL != 0
void puthex6(u_int32 a) {
    puthex(a>>16);
    puthex(a);
}
#endif

int IsValid(u_int32 sector);

extern struct FsPhysical *ph;
extern struct FsMapper *map;
//struct FsMapper *oldMap;


#define NandSelect()   PERIP(GPIO0_CLEAR_MASK) = GPIO0_CS1
#define NandDeselect() PERIP(GPIO0_SET_MASK) = GPIO0_CS1
void NandPutDataAddress(register __c u_int32 addr);
u_int16 NandGetStatus(void);
extern struct FsNandPhys fsNandPhys;


/* Always writes large-page: 2048+64=2112 */
#if 1
//s_int16 LargePageWrite(s_int32 page, u_int16 *data, u_int16 *meta);
#else
s_int16 LargePageWrite(s_int32 page, u_int16 *data, u_int16 *meta) { /*98 words*/
#if DEBUG_LEVEL > -2
    puthex6(page);
    puts("=LPW");
#endif
  PERIP(GPIO0_SET_MASK) = GPIO0_CS1; //NandDeselect(); /* For robustness */
  NandSetWaits(fsNandPhys.waitns);
  PERIP(GPIO0_CLEAR_MASK) = GPIO0_CS1;//  NandSelect();

  NandPutCommand(NAND_OP_READ_A);/*Sets read/write pointer*/
  NandPutCommand(NAND_OP_PREPARE_TO_PROGRAM);
  NandPutDataAddress(page);

  {
      register __c1 u_int16 i;
      register __i3 u_int16 *m = meta;
      register __i1 u_int16 *d = data;
      for (i=0;i<4;i++) {
	  PERIP(NFLSH_DSPIF) = NFLSH_DF_ECC_RESET;
	  NandPutOctets(512, d);
	  m[0] = PERIP(NFLSH_LPL);
	  m[1] = (PERIP(NFLSH_CP_LPH)<<8) | (m[1] & 0xff);
	  NandSwapBad(m);
	  d += 256;
	  m += 8;
      }
  }
  NandPutOctets(4*16,meta);
  /** Must restore the original order! */
  NandSwapBad(meta+0*8);
  NandSwapBad(meta+1*8);
  NandSwapBad(meta+2*8);
  NandSwapBad(meta+3*8);

  NandPutCommand(NAND_OP_PERFORM_PROGRAM); //Program the data and spare
  {
      register u_int16 n = NandGetStatus();
      PERIP(GPIO0_SET_MASK) = GPIO0_CS1;//      NandDeselect();

#if DEBUG_LEVEL != 0
      if (!IsValid(page)) {
	  puts("Fail!");
      }
#endif
      return (n & 1); //(n & 1) ? (s_int16)n : 0; //Error code from flash      
  }
}
#endif


/* Always reads large-page: 2048+64=2112 */
#if 1
s_int16 LargePageRead(s_int32 page, u_int16 *data, u_int16 *meta);
#else
s_int16 LargePageRead(s_int32 page, u_int16 *data, u_int16 *meta) { /*219 words with debugs*/
//  register s_int16 i;
  s_int16 err = 0;
  __y s_int16 reread = 0;
  __y u_int32 ecc[4];
  register u_int16 n;
  register __i3 u_int16 *d = data;
  register __i1 u_int16 *m = meta;

#ifdef PERIP_IN_X
  #define GET_ECC (*(u_int32 *)NFLSH_LPL)
#else
  #define GET_ECC (*(__y u_int32 *)NFLSH_LPL)
#endif
tryAgain:
  d = data;
  m = meta;
  PERIP(GPIO0_SET_MASK) = GPIO0_CS1;//      NandDeselect();
  NandSetWaits(fsNandPhys.waitns);
  PERIP(GPIO0_CLEAR_MASK) = GPIO0_CS1;//  NandSelect();

  NandPutCommand(NAND_OP_READ_A);
  NandPutDataAddress(page & ~3L);
  NandPutCommand(NAND_OP_COMMIT_DATA_ADDRESS);

  {
      register __i2 __y u_int32 *e = &ecc[0];
      for (n=0;n<4;n++) {
	  PERIP(NFLSH_DSPIF) = NFLSH_DF_ECC_RESET;
	  NandGetOctets(512, d);
	  d += 512/2;
	  *e++ = GET_ECC;
      }
  }
  d -= 4*512/2;
  NandGetOctets(4*16, m);

  {
      register __i2 __y u_int32 *e = &ecc[0];
      for (n=0;n<4;n++) {
	  register u_int32 xored;

	  /** swap first and sixth byte (bad block) */
	  NandSwapBad(m);

	  /* e1 e0 e2 05 ff bd ff ff l1 l0 l3 l2 ff ff ff ff  small-page */
	  xored = (((u_int32)(m[1] & 0xff00U)<<8) | m[0]) ^ *e++;
	  if (xored) {
	      register s_int16 oneBits = CountBitsLong(xored);
	      if (++reread <= 3) {
	      again:
#if DEBUG_LEVEL > 0
		  puthex(oneBits);
		  puthex6(page);
		  puts("@err(retry)");
#endif
		  fsNandPhys.waitns += 1; /*!! find the right ws!*/

		  if ((PERIP(NFLSH_CTRL) & 31) < 30)
		      PERIP(NFLSH_CTRL) += 2;   // increase waitstates for retry
		  //PERIP(NFLSH_CTRL)++;   // lcc fails!
		  goto tryAgain;
	      }
	      if (oneBits == 1) {
		  /* Error in ECC, but data is ok: do nothing. */
#if DEBUG_LEVEL > 0
		  puthex6(page);
		  puts("@ECC err (don't worry)");
#endif
	      } else if (oneBits != 12) {
		  /* Fatal error: more than one error bits. */
#if DEBUG_LEVEL > 0
		  puthex(oneBits);
		  puts("@Uncorrectable err (sorry!)");
#endif
		  //goto again;
		  err = -1;
	      } else {
		  /* Correctable error. Correct it! */
		  register int line = NandMingle(xored);
		  register int col = (((line & 1) ^ 1) << 3) | (line >> 9);
#if DEBUG_LEVEL > 0
		  puthex(line & 0x1ff);
		  puthex(col & 7);
		  putchar(' ');
#endif
		  line = (line >> 1) & 0xff;
		  d[line] ^= 1<<col;
#if DEBUG_LEVEL > 0
		  puthex6(page);
		  puts("@Correctable err");
#endif
	      }
	  }
	  d += 512/2;
	  m += 16/2;
      }
  }
  PERIP(GPIO0_SET_MASK) = GPIO0_CS1;//      NandDeselect();
  return err;
}
#endif

#if DEBUG_LEVEL > 2
void PrintData(u_int16 *d, u_int16 l) {
    int i;
    for(i=0;i<l;i++) {
	puthex(*d++);
	if ((i&7) == 7)
	    puts("");
    }
}
#endif


#ifdef USE_MAP
struct FsMapper *FsMapMlcCreate(struct FsPhysical *physical,
				u_int16 cacheSize);
u_int16 FsMapMlcRead(struct FsMapper *map, u_int32 firstBlock,
		     u_int16 blocks, u_int16 *data);
u_int16 FsMapMlcWrite(struct FsMapper *map, u_int32 firstBlock,
		      u_int16 blocks, u_int16 *data);
u_int16 FsMapMlcFlush(struct FsMapper *map, u_int16 flush);

/*const*/ struct FsMapper mlcMapper = {
    0x010c, /*version*/
    256,    /*blocksize in words*/
    0,      /*blocks -- read from CSD*/
    0,      /*cacheBlocks*/
    NULL,//FsMapMlcCreate,
    FsMapFlNullOk,//RamMapperDelete,
    FsMapMlcRead,
    FsMapMlcWrite,
    NULL,//FsMapFlNullOk,//RamMapperFree,
    FsMapMlcFlush,//RamMapperFlush,
    NULL /* no physical */
};

//extern __y u_int16 treeBuild[CACHE_SECTORS][256]; /* cache data */
//extern __y u_int16 g_dcthi[2048];

u_int32 GetPhysical(u_int32 logical) {
    register u_int16 bl = logical / ph->eraseBlockSize;
    register int i = 0;
    while (i < mlc.bads) { /* 15 cycles per loop */
	if (mlc.bad[i] > bl)
	    break;
	bl++;
	i++;
    }
#if DEBUG_LEVEL > 2
    puthex6(logical);
    putchar('>');
    puthex6((logical % ph->eraseBlockSize) + (u_int32)bl * ph->eraseBlockSize);
    puts("");
#endif
    return (logical & (ph->eraseBlockSize-1)) + (u_int32)bl * ph->eraseBlockSize;
}

#if 1
int IsValid(u_int32 sector);
#else
int IsValid(u_int32 sector) {
    /* Always read the spare directly.
       Also works optimally with the physical cache
       (if it isn't written yet, it can be modified)! */
    u_int16 meta[8];
    ph->Read(ph, sector, 1, NULL, meta);
    if ((meta[1] & 0xff) != 0xff /*==0x05*/)
	return 1;
    return 0;
}
#endif

#if 1
u_int16 GetBitArray(register __i0 u_int16 *array, register __a0 int bit);
void SetBitArray(register __i0 u_int16 *array, register __a0 int bit);
void ClearBitArray(register __i0 u_int16 *array, register __a0 int bit);
#else
u_int16 GetBitArray(register __i0 u_int16 *array, register __a0 int bit) {
    register __a1 int t = array[bit >> 4] & (1<<(bit&15));
    return t;
}
void SetBitArray(register __i0 u_int16 *array, register __a0 int bit) {
    array[bit >> 4] |= (1<<(bit&15));
}
void ClearBitArray(register __i0 u_int16 *array, register __a0 int bit) {
    array[bit >> 4] &= ~(1<<(bit&15));
}
#endif

#if 0
struct REPL *GetReplWith(u_int16 block) {
    register int i;
    for (i=0; i<mlc.replaced; i++)
	if ((u_int32)block * ph->eraseBlockSize == mlc.replace[i].new)
	    return &mlc.replace[i];
    return NULL;
}
#endif

void Merge(struct REPL *r);

/* Get the next unused erase block from the replace area. */
u_int16 GetBlock(void) {
    static u_int16 cur = 0;
 re:
    if (cur >= ph->eraseBlocks || cur < (ph->eraseBlocks-RESERVED_AREA))
	cur = (ph->eraseBlocks-RESERVED_AREA);
#if 0
    if (GetReplWith(cur)) {
	cur++;
	goto re;
    }
#else
    {
	register int i;
	for (i=0; i<mlc.replaced; i++) {
	    if ((u_int32)cur * ph->eraseBlockSize == mlc.replace[i].new) {
		cur++;
		goto re;
	    }
	}
    }
#endif

#if DEBUG_LEVEL > 1
    puthex(cur);
    puts(" alloc");
#endif
#if 1 /* Clear it to be certain during debugging. -- does not really slow down. */
    ph->Erase(ph, (u_int32)cur * ph->eraseBlockSize);
    //putch('E');
    //putch('1');
#endif
#if 0
    puthex(cur);
    puts("bl");
#endif
    return cur++;
}

#if 0
struct REPL *PutReplaced(u_int16 block, u_int16 repl) {
    register struct REPL *r = &mlc.replace[mlc.replaced];
    u_int16 *d = r->data;
    memset(r, 0, sizeof(mlc.replace[0]));
    r->data = d;
    //r->bits = 0;
    //r->time = 0;
    r->orig = (u_int32)block * ph->eraseBlockSize;
    r->new  = (u_int32)repl * ph->eraseBlockSize;
    //r->wrPtr = 0;
    replaced++;
    return r;
}
#endif

#if 0
void RemReplaced(struct REPL *r) {
#if 0
    register int i;
    for (i=0; i<mlc.replaced; i++) {
	if (r == &mlc.replace[i]) {
	    u_int16 *d = r->data;
#if DEBUG_LEVEL > 1
	    puthex(i);
	    puthex(mlc.replaced);
	    puthex(mlc.replace[i].orig);
	    puthex(mlc.replace[i].new);
	    puts(" RemReplaced");
#endif
	    memcpy(&mlc.replace[i], &mlc.replace[i+1],
		   sizeof(mlc.replace[0])*(MAXREPL-i));
	    mlc.replaced--;
	    mlc.replace[MAXREPL-1].data = d; /* don't lose the buffer pointer */
	    return;
	}
    }
#else
    memcpy(r, r+1, (char *)&mlc.replace[MAXREPL-1]-(char *)r);
    mlc.replaced--;
#endif
}
#else
#define RemReplaced(r) do{register u_int16 *d = (r)->data;mlc.replaced--;memcpy((r), (r)+1, (char *)&mlc.replace[MAXREPL-1]-(char *)(r));mlc.replace[MAXREPL-1].data=d;}while(0)
#endif

struct REPL *GetReplaced(u_int16 block) {
    register int i;
    for (i=0; i<mlc.replaced; i++)
	if ((u_int32)block * ph->eraseBlockSize == mlc.replace[i].orig)
	    return &mlc.replace[i];
    return NULL;
}


u_int16 dataTmp[2048/2];
__y u_int16 mapTmp[MAX_ERASEBLOCKSIZE/4]; /*Eraseblocksize/4*/

auto void Fill(struct REPL *r) {
    register u_int32 p;
    u_int16 meta[4*8];
    if (r->mask == 15)
	return;
    if (GetBitArray(r->replbitmap, r->se>>2)) {
	register int i;
	for (i=r->wrPtr-4;i>=0;i-=4) {
	    p = r->new + i;
	    ph->Read(ph, p, 1, NULL, meta);
	    if ((meta[4] & (ph->eraseBlockSize-1)) == r->se) {
		/* Found! */
		goto journal;
	    }
	}
#if 0
	puts("fill: no journal");
#endif
    }
    p = r->phys;
 journal:
#if 0
    puthex(r->sector>>16);
    puthex(r->sector);
    //putchar(' ');
    puthex(p>>16);
    puthex(p);
#endif
    /* Use datatmp -- more robust than reading one at a time. */
    LargePageRead(p, dataTmp, meta);
    if (!(r->mask & 1)) {
	memcpy(r->data+0*512/2, dataTmp+0*512/2, 512/2);
#ifdef DIRECT_DEBUG
	puthex(meta[3]);
	puts(" f0");
#endif
    }
    if (!(r->mask & 2)) {
	memcpy(r->data+1*512/2, dataTmp+1*512/2, 512/2);
#ifdef DIRECT_DEBUG
	puthex(meta[3]);
	puts(" f1");
#endif
    }
    if (!(r->mask & 4)) {
	memcpy(r->data+2*512/2, dataTmp+2*512/2, 512/2);
#ifdef DIRECT_DEBUG
	puthex(meta[3]);
	puts(" f2");
#endif
    }
    if (!(r->mask & 8)) {
	memcpy(r->data+3*512/2, dataTmp+3*512/2, 512/2);
#ifdef DIRECT_DEBUG
	puthex(meta[3]);
	puts(" f3");
#endif
    }
    r->mask = 15;
}

auto void SetMeta(register __c s_int32 sector, register __i2 u_int16 *meta) {
    memset(meta, 0xffff, 4*8*sizeof(u_int16));
#if 0
    meta[1] = 0xff05;
    meta[3] = sector;
    meta[4] = sector;
    meta[5] = sector>>16;
    meta[6] = sector;
    sector++;
    meta += 8;
    meta[1] = 0xff05;
//    meta[4] = sector;
//    meta[5] = sector>>16;
    sector++;
    meta += 8;
    meta[1] = 0xff05;
//    meta[4] = sector;
//    meta[5] = sector>>16;
    sector++;
    meta += 8;
    meta[1] = 0xff05;
//    meta[4] = sector;
//    meta[5] = sector>>16;
#else
    meta++;
    *meta = 0xff05;       //[1]
    meta += 2;
    *meta++ = sector; //[3]
    *meta++ = sector;     //[4]
    *meta++ = sector>>16; //[5]
    *meta++ = sector; //[6]
    meta += 2;
    sector++;
    *meta++ = 0xff05;     //[1+8]
    meta += 7;
    sector++;
    *meta++ = 0xff05;     //[1+16]
    meta += 7;
    *meta = 0xff05;       //[1+24]
#endif
}

void Merge(struct REPL *r) {
    /* First get an extra erase block (does not really allocate it). */
    register int i;
    register u_int16 bl = 0;
    int cnt = 0;

//PERIP(GPIO1_CLEAR_MASK) = LED1;

#if 0
    puthex(r->orig>>16);
    puthex(r->orig);
    puts("Merge");
#endif

    bl = GetBlock();

#if DEBUG_LEVEL > 0
    puthex(r->orig);
    putchar(',');
    puthex(r->new);
    puts(" orig,new");

    for (i=0;i<256/16;i++) {
	puthex(r->replbitmap[i]);
    }
    puts("=replbitmap");
#endif

    if (r->mask) {
	Fill(r);
	SetBitArray(r->replbitmap, r->se>>2);
    }

    /* Copy all untouched blocks from r->orig to bl. */
    for (i=0; i<ph->eraseBlockSize; i+=4) {
	u_int16 meta[4*8];
	if ( GetBitArray(r->replbitmap, i>>2) ) {
	    /* replaced data */
	} else {
	    /* copy data that was not overwritten */
	    LargePageRead(r->orig +i, dataTmp, meta);
#if 0
	    if (meta[4] != 0xffff &&
		*(u_int32*)&meta[4] != r->logical+i) {
		putch(0xaa);
		putch(0);
		putch(meta[5]>>8);
		putch(meta[5]);
		putch(meta[4]>>8);
		putch(meta[4]);
		putch(i>>8);
		putch(i);
	    }
#endif
#if 0
	    if (meta[4] != 0xffff &&
		(meta[4] & (ph->eraseBlockSize-1)) != i) {
		putch(meta[4]>>8);
		putch(meta[4]);
		putch(i>>8);
		putch(i);
	    }
#endif
	    //meta[1] = 0xff05;
	    SetMeta(r->logical + i, meta);
	    if (LargePageWrite((u_int32)bl*ph->eraseBlockSize+cnt, dataTmp, meta)) {
		//putch('W');
		//putch('1');
	    }
	    cnt += 4;
	}
    }
    /* Erase r->orig. */
    if (ph->Erase(ph, r->orig)) {
	//putch('E');
	//putch('2');
    }

    memsetY(mapTmp, 0, sizeof(mapTmp));
    for (i=0; i<r->wrPtr;i+=4) {
	u_int16 meta[8];
    reread:
	ph->Read(ph, r->new + i, 1, NULL, meta);
	/* Read error? */
#if 0
	if (meta[4] != meta[6] || meta[4] != meta[3]) {
	    putch(i>>8);
	    putch(i);
	    putch(meta[4]>>8);
	    putch(meta[4]);
	    putch(meta[5]>>8);
	    putch(meta[5]);
	    putch(meta[6]>>8);
	    putch(meta[6]);
	    putch(meta[3]>>8);
	    putch(meta[3]);
	    goto reread;
	}
#endif
	if (CountBitsLong(meta[4] ^ meta[6]) + CountBitsLong(meta[4] ^ meta[3]) > 4)
	    goto reread;
	/*
	  a b c Majority vote
	  1 1 1
	  1 1 0
	  1 0 1
	  0 1 1
	  (a&b) | (a&c) | (b&c) = (a & (b|c)) | (b&c)
	 */
	{
	    register u_int16 ll =
		(meta[4] & (meta[6] | meta[3])) | (meta[6] & meta[3]);
	    ll &= (ph->eraseBlockSize-1);
	    mapTmp[ll >> 2] = i;
	}
    }


    {
	register int j;
	i = 0;
	for (j=0; j<ph->eraseBlockSize; j+=4) {
	    u_int16 meta[4*8];
	    if (GetBitArray(r->replbitmap, j>>2)) {
		if (r->mask && r->se == j) {
		    memcpy(dataTmp, r->data, 2048/2);
		    SetMeta(r->sector, meta);
		} else {
		    /*Find the page from journal.
		      Use a cache for speedup. */
		    LargePageRead(r->new + mapTmp[j>>2], dataTmp, meta);
#if 0
		    if (meta[4] != 0xffff &&
			*(u_int32*)&meta[4] != r->logical+j) {
			putch(0xaa);
			putch(1);
			putch(meta[5]>>8);
			putch(meta[5]);
			putch(meta[4]>>8);
			putch(meta[4]);
			putch(j>>8);
			putch(j);
		    }
#endif
		}
	    } else {
		LargePageRead((u_int32)bl * ph->eraseBlockSize + i, dataTmp, meta);
#if 0
		if (meta[4] != 0xffff &&
		    *(u_int32*)&meta[4] != r->logical+j) {
		    putch(0xaa);
		    putch(2);
		    putch(meta[5]>>8);
		    putch(meta[5]);
		    putch(meta[4]>>8);
		    putch(meta[4]);
		    putch(j>>8);
		    putch(j);
		}
#endif
		i += 4;
	    }
	    //SetMeta(r->orig + j, meta);
	    if (LargePageWrite(r->orig + j, dataTmp, meta)) {
		//putch('W');
		//putch('5');
	    }
	}
    }
    /* Erase r->new */
    if (ph->Erase(ph, r->new)) {
	//putch('E');
	//putch('3');
    }

    /* Erase bl. */
    if (ph->Erase(ph, (u_int32)bl * ph->eraseBlockSize)) {
	//putch('E');
	//putch('4');
    }

    RemReplaced(r);
//PERIP(GPIO1_SET_MASK) = LED1;
}


u_int16 FsMapMlcRead(struct FsMapper *map, u_int32 sector,
		     u_int16 blocks, u_int16 *data) {
    u_int16 bc = 0;
#ifdef USE_USB
    while (bc < blocks) {
	register u_int32 phy = GetPhysical(mlc.dataStart + sector);
	register u_int16 se = phy & (ph->eraseBlockSize-1);
	u_int16 bl = phy / ph->eraseBlockSize;
	register struct REPL *r = GetReplaced(bl);
	u_int16 meta[8];
#if 0
	puthex(sector);
	puts("=R");
#endif


#if DEBUG_LEVEL > 1
	//puthex6(sector);
	//putchar(' ');
	puthex6(phy);
	puts(" RD");
#endif
	if (r) {
	    /* Replaced, check if a new version exists. */
#if 0 //DEBUG_LEVEL > 1
	    puts("r");
#endif
	    if (r->se == (se & ~3) && (r->mask & (1<<(se & 3)))) {
		/* Got it in cache. */
#if 0
	      puthex(phy>>16);
	      puthex(phy);
		puts(" ch");
#endif
#if DEBUG_LEVEL > 2
		printf(" read ch %06lx %d\n", phy, r->data[(phy&3)*8+7]);
#endif
		memcpy(data, &r->data[(phy & 3)*512/2], 512/2*sizeof(u_int16));
		meta[1] = 0xff05;
		goto noread;
	    } else if ( GetBitArray(r->replbitmap, se>>2) ) {
		/* a new version, try to locate it. */
		register int i;
#if DEBUG_LEVEL > 1
		puts(" yes");
#endif
		for (i=r->wrPtr-4; i>=0; i-=4) {
		    register u_int16 ll;
		    ph->Read(ph, r->new + i, 1, NULL, meta);
		    ll = (meta[4] & (meta[6] | meta[3])) | (meta[6] & meta[3]);

		    if ((se & ~3) == (ll & (ph->eraseBlockSize-1))) {
			phy = r->new + i + (se&3);
#if DEBUG_LEVEL > 1
			//puthex(r->wrPtr-i);
			//putchar(' ');
			puthex6(phy);
			puts(" repl");
#endif
#if 0
			//puthex(phy);
			puts("j");
#endif
			break;
		    }
		}
#if 0
		if (i<0) {
		  puthex(phy>>16);
		  puthex(phy);
		  puts(" no j!");
		}
#endif
		/* Error, replbitmap invalid */
	    } else {
		/* Not replaced, use original */
	    }
	} else {
	    /* Not replaced, use original if valid.. */
	}
	ph->Read(ph, phy, 1, data, meta);
    noread:
	if ((meta[1] & 0xff) != 0x05) {
	    /* Not a data block. Sector not available or never written. */
	    /* Return ZERO data! */
#if DEBUG_LEVEL > 0
	    puthex(sector);
	    puts("NA");
#endif
#if 0
	    puts("NA");
#endif
	    memset(data, 0, 512/2);
#if 0
	} else if (meta[4] == (u_int16)sector &&
		   meta[5] == (u_int16)(sector>>16)) {
	    /* data ok */
	} else {
#if DEBUG_LEVEL > 0
	    puthex(meta[5]);
	    puthex(meta[4]);
	    putchar('!');
	    puthex(sector);
	    puts("");
#endif
	    /* Sector not found at the right place. */
#endif
	}
gotit:
	data += 512/2;
	sector++;
	bc++;
    }
#endif
    return bc;
}
u_int16 FsMapMlcWrite(struct FsMapper *map, u_int32 sector,
		      u_int16 blocks, u_int16 *data) {
    u_int16 bc = 0;
    while (bc < blocks) {
	register u_int32 phy = GetPhysical(mlc.dataStart + sector);
	struct REPL *r;
	u_int16 se;
#if 0
	puthex(sector);
	puts("=W");
#endif
#if DEBUG_LEVEL > 1
	puthex6(sector);
	puts(" WR");
#endif
//	data[1] = 0xff05;
//	data[4] = sector;
//	data[5] = sector>>16;

    re:
	se = phy % ph->eraseBlockSize;
	r = GetReplaced(phy / ph->eraseBlockSize);
	if (!r) {
	    /*AllocReplace*/
	    if (mlc.replaced >= MAXREPL) { /* if all slots used */
		register int i;
		register u_int16 idx = 0, high = 0U;
#if DEBUG_LEVEL > 0
		puts("Out of replace space, force flush MERGE");
#endif
		for (i=0; i<mlc.replaced; i++) {
		    register u_int16 b = mlc.replace[i].bits + mlc.replace[i].time *1;
		    if (high < b) {
			high = b;
			idx = i;
		    }
		}
		Merge(&mlc.replace[idx]);
	    }
	    {
		register int i;
		for (i=0;i<MAXREPL;i++)
		    if (mlc.replace[i].time < 0xffffU)
			mlc.replace[i].time++;
	    }
	    //r = PutReplaced(phy / ph->eraseBlockSize, GetBlock());
	    r = &mlc.replace[mlc.replaced];
	    {
		register u_int16 *d = r->data;
		memset(r, 0, sizeof(*r));
		r->data = d;
	    }
	    r->inOrder = 1; /* ordered by default */
	    r->logical = sector & ~(u_int32)(ph->eraseBlockSize-1);
	    r->orig    = phy    & ~(u_int32)(ph->eraseBlockSize-1);
	    r->new     = (u_int32)GetBlock() * ph->eraseBlockSize;
	    mlc.replaced++;
	}
	if (r->se != (se & ~3) && r->mask) {
	    /* find the unwritten blocks from r->old or journal */
	    u_int16 meta[4*8];
	    Fill(r);
#if DEBUG_LEVEL > 0
	    //printf(" %04x %06lx\n", r->orig, r->sector);
#endif
#ifdef DIRECT_WRITE
	    if (r->wrPtr == 0 && /* Stop writing if any block goes to journal. */
		r->se == r->wrDirect && (r->se != 0 || !IsValid(r->orig))) {
		/* Write to the original sector directly. */
putch('D');
		SetMeta(r->sector, meta);
		LargePageWrite(r->sector, r->data, meta);
		r->wrDirect += 4;
	    } else
#endif /*DIRECT_WRITE*/
	    {
		if (r->wrPtr >= ph->eraseBlockSize) {
#if DEBUG_LEVEL > 0
		    puts(" Journal for block full!");
#endif
		    Merge(r);
		    goto re;
		}
		/* Write to journal */
		SetMeta(r->sector, meta);
		if (LargePageWrite(r->new + r->wrPtr, r->data, meta)) {
		    //putch('W');
		    //putch('6');
		}
		if (r->se != r->wrPtr)
		    r->inOrder = 0; /* no longer ordered */

		r->wrPtr += 4;
		/* replbitmap can be reduced to 1/4 */
		if (!GetBitArray(r->replbitmap, r->se>>2))
		    r->bits++;
		SetBitArray(r->replbitmap, r->se>>2);
	    }
	    r->mask = 0;
	}
	/* the same 2k */
#if DEBUG_LEVEL > 0
	//printf("->c %06lx %d\n", phy, r->bits);
#endif
	r->sector = sector & ~3L;
	r->phys = phy & ~3L;
	r->se = se & ~3;
//Fill(r);
	r->mask |= (1 << (se & 3));
	memcpy(&r->data[(se & 3)*512/2], data, sizeof(u_int16)*512/2);

/*SHOULD BE ACTIVE! Speeds up by flushing "terminated" blocks. */
	if (r->mask == 15 &&
	    r->se + 4 == ph->eraseBlockSize ) {
#ifdef DIRECT_WRITE
	    if (r->wrPtr == 0 && r->se == r->wrDirect) {
		/* Write to the original sector */
		u_int16 meta[4*8];
putch('F');
		SetMeta(r->sector, meta);
		LargePageWrite(r->sector, r->data, meta);
		RemReplaced(r);
	    } else
#endif /*DIRECT_WRITE*/
	    {
#if DEBUG_LEVEL > 0
		puts("MERGE filled");
#endif
//putch('M');
		Merge(r);
	    }
	}

	data += 512/2;
	sector++;
	bc++;
    }
    return bc;
}
u_int16 FsMapMlcFlush(struct FsMapper *map, u_int16 hard) {
#if 0
    puts("=F");
#endif
#if DEBUG_LEVEL > 0
    puthex(replaced);
    puts("Flush()");
#endif
    /*Note: this slows down writes by cygwin cp considerably: 272->121kB/sec. */
    while (mlc.replaced) {
	Merge(&mlc.replace[0]);
    }
    return 1;
}
#if 0
/* In checkmlc.c */
struct FsMapper *FsMapMlcCreate(struct FsPhysical *physical,
				u_int16 cacheSize) {
    /* Scan erase blocks to locate the first data block,
       bad blocks, and possible replaced blocks. */
    ...
    return &mlcMapper;
}
#endif

#endif/*USE_MAP*/



#ifdef USE_USB
#include <usblowlib.h>
#endif/*USE_USB*/

#ifdef PATCH_SCSI_COMMAND
auto u_int16 NewDiskProtocolCommand(register __i2 u_int16 *cmd) {
#if 1
    if ((cmd[OPERATION_CODE] & 0xff) == SCSI_START_STOP_UNIT) {
	/* 0:OP 1:imm0 2:res 3:res 4:power7-4 res3-2 loej1 start0 5:control */
	startUnit = cmd[4] & 3;
    } else if (startUnit == 2) {
	/* The next command after STOP UNIT will not be serviced
	   fully because we will turn ourselves off. */
	startUnit = 4;
    } else
#endif
#if 0
    if ((cmd[OPERATION_CODE] & 0xff) == SCSI_INQUIRY &&
	(cmd[1/*flags__pageCode*/] & 0x01ff) == 0x0180) {
	//Extended page: Serial Number page requested, See SCSI2 chapter 8.3.4.5
	/*Periph qual,type | page code, Res | Len of sernum in bytes, "10001005" */
	static const __y u_int16 testSerNumber[] = {
	    0x0080, 0x0008, 0x3130, 0x3030, 0x3130, 0x3035
	};
	SCSI.DataOutBuffer = SCSI.DataBuffer;
	//SCSI.Status = SCSI_OK;
	SCSI.State = SCSI_DATA_TO_HOST;
	memcpyYX(SCSI.DataBuffer, testSerNumber, 6);
	SCSI.DataOutSize = 12;             
	return 0; /* handled */
    }
#endif
    return cmd[OPERATION_CODE]; /* perform the command */
}
#endif



// ---- USB DESCRIPTOR STUFF ---- //

#define DT_LANGUAGES 0
#define DT_VENDOR 1
#define DT_MODEL 2
#define DT_SERIAL 3
#define DT_DEVICE 4
#define DT_CONFIGURATION 5

#if 0
#define VENDOR_NAME_LENGTH 7
const u_int16 myVendorNameStr[] = {
  ((VENDOR_NAME_LENGTH * 2 + 2)<<8) | 0x03,
  'V'<<8,'l'<<8,'s'<<8,'i'<<8,'F'<<8,'i'<<8,'n'<<8};
#endif

#define MODEL_NAME_LENGTH 8
const u_int16 myModelNameStr[] = {
  ((MODEL_NAME_LENGTH * 2 + 2)<<8) | 0x03,
  'M'<<8,'L'<<8,'C'<<8,'A'<<8,'u'<<8,'d'<<8,'i'<<8,'o'<<8};

#define SERIAL_NUMBER_LENGTH 12
u_int16 mySerialNumberStr[] = {
  ((SERIAL_NUMBER_LENGTH * 2 + 2)<<8) | 0x03,
  '0'<<8, // You should put your own serial number here
  '0'<<8,
  '0'<<8,
  '0'<<8,
  '0'<<8,
  '0'<<8,
  '0'<<8,
  '0'<<8,
  '0'<<8,
  '0'<<8,
  '0'<<8,
  '1'<<8, // Serial number should be unique for each unit
};

// This is the new Device Descriptor. See the USB specification! 
const u_int16  myDeviceDescriptor [] = "\p"
  "\x12" // Length
  "\x01" // Type (Device Descriptor)
  "\x10" // LO(bcd USB Specification version) (.10)
  "\x01" // HI(bcd USB Specification version) (1.)
  "\x00" // Device Class (will be specified in configuration descriptor)
  "\x00" // Device Subclass
  "\x00" // Device Protocol
  "\x40" // Endpoint 0 size (64 bytes)

  "\xfb" // LO(Vendor ID) (0x19fb=VLSI Solution Oy)
  "\x19" // HI(Vendor ID)

  "\xf3" // LO(Product ID) (0xeee0 - 0xeeef : VLSI Customer Testing)
  "\xee" // HI(Product ID) (customers can request an ID from VLSI)

  "\x00" // LO(Release Number)
  "\x00" // HI(Release Number)
  "\x01" // Index of Vendor (manufacturer) name string
  "\x02" // Index of Product (model) name string (most shown by windows)
  "\x03" // Index of serial number string
  "\x01" // Number of configurations (1)
;

#define CONFIG_DESC_SIZE 123

u_int16 myConfigurationDescriptor[] = "\p"
  
  // ______Configuration_descriptor____ at offset 0
  "\x09"   // Length: 9 bytes // (9)
  "\x02"   // Descriptor type: Configuration // (2)
  "\x7b"    // LO(TotalLength) (123)
  "\x00"    // HI(TotalLength) (0)
  "\x03"    // Number of Interfaces (3)
  "\x01"    // bConfigurationValue (1)
  "\x00"    // iConfiguration (0)
  "\x80"    // Attributes (128)
  "\x32"    // Max. Power 100mA (50)
  

  // ______Interface_descriptor____ at offset 9
  "\x09"   // Length: 9 bytes // (9)
  "\x04"   // Descriptor type: Interface // (4)
  "\x00"   // Interface number  // (1)
  "\x00"   // Alternate setting  // (0)
  "\x00"   // Number of endpoints  // (0)
  "\x01"   // Interface Class: Audio // (1)
  "\x01"   // Interface Subclass: Audio Control // (1)
  "\x00"   // Interface Protocol  // (0)
  "\x00"   // String index  // (0)
  
  // ______Audio Control Class Interface_descriptor____ at offset 18
  "\x09"   // Length: 9 bytes // (9)
  "\x24"   // Descriptor type: Audio Control Class Interface // (36)
  "\x01"   // Subtype Header // (1)
  "\x00"    // (0)
  "\x01"    // (1)
  "\x1E"    // (30)
  "\x00"    // (0)
  "\x01"    // (1)
  "\x01"    // (1)
  
  // ______Audio Control Class Interface_descriptor____ at offset 27
  "\x0C"   // Length: 12 bytes // (12)
  "\x24"   // Descriptor type: Audio Control Class Interface // (36)
  "\x02"   // Subtype Input Terminal // (2)
  "\x01"    // (1)
  "\x01"    // (1)
  "\x01"    // (1)
  "\x00"    // (0)
  "\x02"    // (2)
  "\x00"    // (0)
  "\x00"    // (0)
  "\x00"    // (0)
  "\x00"    // (0)
  
  // ______Audio Control Class Interface_descriptor____ at offset 39
  "\x09"   // Length: 9 bytes // (9)
  "\x24"   // Descriptor type: Audio Control Class Interface // (36)
  "\x03"   // Subtype Output Terminal // (3)
  "\x02"    // (2)
  "\x04"    // (4)
  "\x03"    // (3)
  "\x00"    // (0)
  "\x01"    // (1)
  "\x00"    // (0)
  
  // ______Interface_descriptor____ at offset 48
  "\x09"   // Length: 9 bytes // (9)
  "\x04"   // Descriptor type: Interface // (4)
  "\x01"   // Interface number  // (1)
  "\x00"   // Alternate setting  // (0)
  "\x00"   // Number of endpoints  // (0)
  "\x01"   // Interface Class: Audio // (1)
  "\x02"   // Interface Subclass: Audio Streaming // (2)
  "\x00"   // Interface Protocol  // (0)
  "\x00"   // String index  // (0)
  
  // ______Interface_descriptor____ at offset 57
  "\x09"   // Length: 9 bytes // (9)
  "\x04"   // Descriptor type: Interface // (4)
  "\x01"   // Interface number  // (1)
  "\x01"   // Alternate setting  // (1)
  "\x01"   // Number of endpoints  // (1)
  "\x01"   // Interface Class: Audio // (1)
  "\x02"   // Interface Subclass: Audio Streaming // (2)
  "\x00"   // Interface Protocol  // (0)
  "\x00"   // String index  // (0)
  
  // ______Audio Streaming Class Interface_descriptor____ at offset 66
  "\x07"   // Length: 7 bytes // (7)
  "\x24"   // Descriptor type: Audio Streaming Class Interface // (36)
  "\x01"   // Subtype  // (1)
  "\x01"    // (1)
  "\x0C"    // (12)
  "\x01"    // (1)
  "\x00"    // (0)
  
  // ______Audio Streaming Class Interface_descriptor____ at offset 73
  "\x0B"   // Length: 11 bytes // (11)
  "\x24"   // Descriptor type: Audio Streaming Class Interface // (36)
  "\x02"   // Subtype  // (2)
  "\x01"    // (1)
  "\x02"    // (2)
  "\x02"    // (2)
  "\x10"    // (16)
  "\x01"    // (1)
  "\x44"    // (68)
  "\xAC"    // (172)
  "\x00"    // (0)
  
  // ______Endpoint_descriptor____ at offset 84
  "\x09"   // Length: 9 bytes // (9)
  "\x05"   // Descriptor type: Endpoint // (5)
  "\x01"   // Endpoint Address  // (1)
  "\x09"   // Attributes.TransferType Isochronous(1) // (9)
  "\xFF"   //EP Size LSB // (255)
  "\x03"   //EP Size MSB (total 1023 bytes) // (3)
  "\x01"   //Polling Interval ms // (1)
  "\x00"   //Refresh // (0)
  "\x00"   //Sync Address // (0)
  
  // ______Class Endpoint_descriptor____ at offset 93
  "\x07"   // Length: 7 bytes // (7)
  "\x25"   // Descriptor type: Class Endpoint // (37)
  "\x01"   // Subtype  // (1)
  "\x00"    // (0)
  "\x00"    // (0)
  "\x00"    // (0)
  "\x00"    // (0)

  // ______Interface_descriptor____ at offset 9
  "\x09"   // Length: 9 bytes // (9)
  "\x04"   // Descriptor type: Interface // (4)
  "\x02"   // Interface number  // (2)
  "\x00"   // Alternate setting  // (0)
  "\x02"   // Number of endpoints  // (2)
  "\x08"   // Interface Class: Mass Storage // (8)
  "\x06"   // Interface Subclass:  // (6)
  "\x50"   // Interface Protocol  // (80)
  "\x00"   // String index  // (0)
  
  // ______Endpoint_descriptor____ at offset 18
  "\x07"   // Length: 7 bytes // (7)
  "\x05"   // Descriptor type: Endpoint // (5)
  "\x82"   // Endpoint Address  // (130)
  "\x02"   // Attributes.TransferType Bulk(2) // (2)
  "\x40"   //EP Size LSB // (64)
  "\x00"   //EP Size MSB (total 64 bytes) // (0)
  "\x10"   //Polling Interval ms // (16)
  
  // ______Endpoint_descriptor____ at offset 25
  "\x07"   // Length: 7 bytes // (7)
  "\x05"   // Descriptor type: Endpoint // (5)
  "\x03"   // Endpoint Address  // (3)
  "\x02"   // Attributes.TransferType Bulk(2) // (2)
  "\x40"   //EP Size LSB // (64)
  "\x00"   //EP Size MSB (total 64 bytes) // (0)
  "\x10"   //Polling Interval ms // (16)
  // end of descriptor  
  ;



// Get Packet from USB receive ring buffer

// Interrupt handler for USB perip
auto void Interrupt0() {
  static u_int16 audioStash[91];
  register u_int16 g = PERIP(USB_STATUS);
  register u_int16 st;
  //SCOPE_PUT(g>>8);

PERIP(USB_EP_ST3) |= USB_STF_IN_FORCE_NACK;

  if (PERIP(USB_RDPTR) != PERIP(USB_WRPTR)) {
    st = USEX(PERIP(USB_RDPTR) + USB_RECV_MEM);
    if (((st >> 10) & 0x0f) == AUDIO_ISOC_OUT_EP) {

      register __b0 u_int16 tmp;
      register __b1 u_int16 lenW;
      /*
       * There should be now some data waiting at receive buffer
       */
      lenW = ((st & 0x3ff) + 1)>>1;
      tmp = (PERIP(USB_RDPTR) + 1) & 0x03FF; 
      if (lenW <= (ENDPOINT_SIZE_1/*+1*/)>>1) {
	RingBufCopyX(audioStash, (void *)(tmp + USB_RECV_MEM), lenW);
      }
      PERIP(USB_RDPTR) = (lenW + tmp) & 0x03FF;

      if (AudioBufFree() > 45) {
	Enable();
	AudioPacketFromUSB(audioStash, lenW);
	Disable();
      }
    } else {
      /* not audio packet */
    }
  }

  if (g & USB_STF_SOF) {
    PERIP(USB_STATUS) = USB_STF_SOF;
    USB.lastSofTime = ReadTimeCount();
    USB.lastSofFill = AudioBufFill();
  }
}


void ResetBulkEndpoints(void);
// New USB handler
void MyUSBHandler(){
  __y u_int16 ep;
  u_int16 g;
  register u_int16 info;

  if ((s_int16)PERIP(USB_STATUS) < 0) {
    InitUSB(0);
    PERIP(USB_EP_ST3) &= ~USB_STF_IN_FORCE_NACK; // Release EP3
  }
  
  if (USB.EPReady[0] == 0) {
    USBContinueTransmission(0);
  }
  if (USB.EPReady[MSC_BULK_IN_ENDPOINT] == 0) {
    USBContinueTransmission(MSC_BULK_IN_ENDPOINT);
  }

  // Are there packet(s) in the receive buffer?
  Disable();
  if (PERIP(USB_RDPTR) != PERIP(USB_WRPTR) &&
      (USEX(PERIP(USB_RDPTR) + USB_RECV_MEM) & 0x3c00)
      != (AUDIO_ISOC_OUT_EP << 10) ) {
    Enable();
    PERIP(USB_EP_ST3) |= USB_STF_IN_FORCE_NACK;
    ep = USBReceivePacket(&USB.pkt);

    if (USB.pkt.length) { //Does it actually contain data?
      if (ep == 0) { // Setup Packet

	// NOTE: this function is timing critical!
	//EP0 shalt not be transmitting now.
	USB.EPReady[0] = 1; //EP0 shalt be ready now.
    
	if ((s_int16)USB.pkt.payload[0] >= 0 &&
	    (USB.pkt.payload[USB_SP_REQUEST] & 0xff) ==
	    USB_REQUEST_SET_INTERFACE) {
	  USBResetEndpoint(0);
	  /* index = interface, value = alternate setting */
	  USB.interfaces =
	    (USB.pkt.payload[USB_SP_VALUE] & 0xff00U) |
	    (USB.pkt.payload[USB_SP_INDEX] >> 8);
#if 1 /* ok with windoze */
	  if ( USB.pkt.payload[USB_SP_INDEX] != 0x0100 /*0x0001*/)
	    ResetBulkEndpoints();
#endif
	  USBSendZeroLengthPacketToEndpoint0(); 
	} else {
	  RealDecodeSetupPacket();
	}

      } else if (ep == AUDIO_ISOC_OUT_EP) {
	AudioPacketFromUSB(USB.pkt.payload, (USB.pkt.length/*+1*/)/2);
      } else if (ep == MSC_BULK_OUT_ENDPOINT) {
	RealMSCPacketFromPC(&USB.pkt);
	PERIP(USB_EP_ST3) &= ~USB_STF_IN_FORCE_NACK;
      } else {
	// Unknown packet
      }
    } else {
      // Empty packet
    }
  } else {
    Enable();
  }
  PERIP(USB_EP_ST3) &= ~USB_STF_IN_FORCE_NACK;
}


#define SWITCHED_KEYS
#ifdef SWITCHED_KEYS
/*
  SPECIAL KEY ORDER FOR Hakuto!
  D0 - VOL+        VOL+
  D1 - VOL-        VOL-
  D2 - PREV        PREV
  D3 - PLAY/PAUSE  PLAY/PAUSE
  D4 - NEXT        IC 25PIN LED ON/OFF KEY(L- H?L?H).
  POWERIN - POWER ON  NEXT/FF 

 */
const struct KeyMapping myKeyMap[] = {
  {KEY_2, ke_volumeDown},
  {KEY_LONG_PRESS|KEY_2, ke_volumeDown},
  {KEY_1, ke_volumeUp},
  {KEY_LONG_PRESS|KEY_1, ke_volumeUp},
//  {KEY_LONG_PRESS|KEY_POWER, ke_powerOff},
  {0, ke_null}
};

#endif/*SWITCHED_KEYS*/

void mlc_structure_too_large_error(void);
void main(void) {
    //ph = FsPhNandCreate(1);
//    int i;

#pragma msg 36 off
    if (sizeof(mlc) > 512)
	mlc_structure_too_large_error();
#pragma msg 36 on

#if 0
    clockX = 7;
    SetRate(100);
#endif

#if 0 /*done in checkmlc */
    /* Set the leds after nand-boot! */
    PERIP(GPIO1_SET_MASK) =  LED1;      /* POWER led on */
    PERIP(GPIO1_DDR)   |=  LED1|LED2; /* SI and SO to outputs */
    PERIP(GPIO1_MODE)  &= ~(LED1|LED2); /* SI and SO to GPIO */
    ph = FsPhNandCreate(1); /* Init the default physical first. */
#endif


//fsNandPhys.waitns = 50;
#ifdef USE_MAP
    map = &mlcMapper; //FsMapMlcCreate(ph, 0); /* Then create mapper v2. */
    mlcMapper.blocks = mlc.dataBlocks;
#else
    map = FsMapTnCreate(ph, 0); /* tiny mapper */
#endif




#ifdef SWITCHED_KEYS
    currentKeyMap = myKeyMap;
#endif
#if 1
    //PlayerVolume();

  // voltages when USB is attached and active
  //  voltages[voltCoreUSB] = 31;     // 31=max
  //  PERIP(GPIO1_ODATA) |=  LED1|LED2;      /* POWER led on */
    //memset(epPackets,0,sizeof(epPackets));

  // Set Hook Functions
    //SetHookFunction((u_int16)USBHandler, MyUSBHandler);
    //SetHookFunction((u_int16)DecodeSetupPacket, MyDecodeSetupPacket);  
    //SetHookFunction((u_int16)MSCPacketFromPC, MyMSCPacketFromPC);
    //SetHookFunction((u_int16)InitUSBDescriptors, MyInitUSBDescriptors);

    // Install USB Interrupt Handler
    WriteIRam(0x20+INTV_USB, ReadIRam((u_int16)InterruptStub0));
    //PERIP(INT_ENABLEL) |= INTF_USB;
    PERIP(INT_ENABLEL) = INTF_RX | INTF_TIM0 | INTF_USB;
#endif





#ifdef USE_USB
//    while(1)
    if (USBIsAttached()) {
#if DEBUG_LEVEL > 0
	puts("Going to MassStorage");
#endif
	PERIP(GPIO1_SET_MASK) = LED2;      /* feature LED on */
#ifdef PATCH_SCSI_COMMAND
	SetHookFunction((u_int16)MSCPacketFromPC, PatchDiskProtocolCommand);
#endif
	/*
	  We replace the MassStorage() functionality.
	*/
#if 0 /*done in checkmlc */
	voltages[voltIoUSB] = 31; /* 3.65V */
	voltages[voltAnaUSB] = 31; /* 3.65V */
	PowerSetVoltages(&voltages[voltCoreUSB]);
	/** a 2.5ms-10ms delay to let the voltages change (rise) */
	BusyWait10();
#endif
	LoadCheck(NULL, 1); /* 4.0x clock required! */
	InitUSB(USB_MASS_STORAGE);

#if 1
	//USB.descriptorTable[DT_VENDOR] = myVendorNameStr;
	USB.descriptorTable[DT_MODEL]  = myModelNameStr;
	USB.descriptorTable[DT_SERIAL] = mySerialNumberStr;
	USB.descriptorTable[DT_DEVICE] = myDeviceDescriptor;
	USB.descriptorTable[DT_CONFIGURATION] = myConfigurationDescriptor;
	USB.configurationDescriptorSize = CONFIG_DESC_SIZE;
#endif

	while (1) {
	    MyUSBHandler();
#if 1
	    PERIP(USB_CONTROL) |= USB_STF_SOF | USB_STF_RX;
	    ScsiTaskHandler();    
#endif

	    /* If no SOF in such a long time.. */
	    if (USBWantsSuspend()) {
		/* Do we seem to be detached?
		   Note: Now also requires USBWantsSuspend() to return TRUE! */
		if (USBIsDetached()) {
		    break;
		}
#if 0 /* Suspend */
		{
		    register u_int16 io = PERIP(GPIO1_ODATA);
		    PERIP(GPIO1_ODATA) = io & ~(LED1|LED2);      /* LEDs off */
		    PERIP(SCI_STATUS) |= SCISTF_ANA_PDOWN;       /*fix suspend*/
		    USBSuspend(0); /* no timeout */
		    PERIP(GPIO1_ODATA) = io;      /* LEDs back to old states */
		    /* with USB audio the LED settings are handled by user interface */
		}
		USB.lastSofTime = ReadTimeCount();
#endif
#if 0 /*panze version*/
		{
		  register u_int16 io = PERIP(GPIO1_ODATA);
		  PERIP(GPIO1_ODATA) = io & ~(LED1|LED2);      /* LED's off */
		  USBSuspend(0); /* no timeout */
		  PERIP(GPIO1_ODATA) = io;      /* Power LED on */
		  /* with USB audio the LED settings are handled by user interface */
		}
		USB.lastSofTime = ReadTimeCount();
#endif

	    }
#if 0 /*should be 1 for volume control to work (when no audio data)!*/
	    if (AudioBufFill() < 32) {
	      memset(tmpBuf, 0, sizeof(tmpBuf));
	      AudioOutputSamples(tmpBuf, sizeof(tmpBuf)/2);
	      IdleHook(); /* calls the user interface -- handle volume control */
	    }
#endif
#if 1
	    /* Handle volume control */
	    //NandPutCommand(0);
	    IdleHook();
#endif

#ifdef PATCH_SCSI_COMMAND
	    if (startUnit == 4) {
		break;
	    }
#endif
	}
	PERIP(SCI_STATUS) &= ~SCISTF_USB_PULLUP_ENA; /* Take us 'out of the bus' */
	map->Flush(map, 2);
	PERIP(USB_CONFIG) = 0x8000U; /* Reset USB */

#ifdef PATCH_SCSI_COMMAND
	if (startUnit == 4) {
	    PowerOff();
	}
#endif

#if 0
	/* Set player voltages. No need to wait for the voltages to drop. */
	LoadCheck(NULL, 0);
	PowerSetVoltages(&voltages[voltCorePlayer]);
#endif

#ifdef PATCH_SCSI_COMMAND
	// SetHookFunction((u_int16)MSCPacketFromPC, RealMSCPacketFromPC);
#endif
	PERIP(GPIO1_CLEAR_MASK) = LED2;      /* LED off */
    }
#endif

#ifdef USE_MAP
    map = FsMapTnCreate(ph, 0); /* tiny mapper */
#endif

#ifdef SWITCHED_KEYS
    //currentKeyMap = sixKeyMap; /*restored in checkmlc*/
#endif
#define PROG_START 0 //12288 //8192
    ph->Read(ph, PROG_START/512, 16, mallocAreaX, minifatBuffer);
    while (1) {
	BootFromX(mallocAreaX+8);
    }
}
