#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <vs1000.h>

#include <player.h>
#include <mappertiny.h>
#include <minifat.h>
#include <vsNand.h>

//#define DUMP_LARGE_PAGE
#define DUMP_PHYS
//#define DUMP_FAT
//#define DUMP_DIR
//#define WR_SECTOR
//#define MAP_FMT
//#define WR_IMAGE
//#define WR_BITS
//#define MARK_BAD

extern struct FsNandPhys fsNandPhys;
struct FmfMeta {
  u_int16 ecc01;
  u_int16 ecc2AndType;
  u_int16 reservedAndBadBlock;
  u_int16 unused;
  u_int32 logicalPageNo;	/* For root node, this is last used */
  s_int32 newBranch;		/* Page # for root node, non- -1 for others */
};
#define FS_MAP_NON_FULL 4

#define FMF_TYPE_INVALIDATED	  0
#define FMF_TYPE_ROOT		  1
#define FMF_TYPE_BRANCH2	  2
#define FMF_TYPE_BRANCH3	  3
#define FMF_TYPE_BRANCH4	  4
#define FMF_TYPE_DATA		  5
#define FMF_TYPE_UNUSED		255

#define FMF_BRANCH_LEVELS	4
#define FMF_EMPTY_BLOCKS	2



extern struct FsPhysical *ph;
extern struct FsMapper *map;
extern struct Codec *cod;
//extern struct CodecServices cs;

__y const char hex[] = "0123456789abcdef";
void puthex(u_int16 a) {
  char tmp[8];
  tmp[0] = hex[(a>>12)&15];
  tmp[1] = hex[(a>>8)&15];
  tmp[2] = hex[(a>>4)&15];
  tmp[3] = hex[(a>>0)&15];
  tmp[4] = ' ';
  tmp[5] = '\0';
  fputs(tmp, stdout);
}
void fputhex6(u_int32 a, FILE *fp) {
  char tmp[8];
  tmp[0] = hex[(a>>20)&15];
  tmp[1] = hex[(a>>16)&15];
  tmp[2] = hex[(a>>12)&15];
  tmp[3] = hex[(a>>8)&15];
  tmp[4] = hex[(a>>4)&15];
  tmp[5] = hex[(a>>0)&15];
  tmp[6] = '\0';
  fputs(tmp, fp);
}
void puthex6(u_int32 a) {
    fputhex6(a, stdout);
}
#define NAN 0x80000000L
long GetParam(char *str) {
  static char in[200];
  char *tmp;
  register int i;
  long value;
  if (str) {
    fputs(str, stdout);
    fflush(stdout);
  }
  fgets(in, 200, stdin);
  i = strlen(in);
  while (i && (in[i-1] == '\n' || in[i-1] == '\r'))
    in[--i] = '\0';
  value = strtol(in, &tmp, 0);
  if (!in[0] || (tmp && *tmp && *tmp != '\n' && *tmp != '\r'))
    return NAN;
  return value;
}
void PrintSector(u_int16 *data, u_int16 *spare) {
#if 1
  register int i, b = 0;
  if (data) {
    puts("d");
    for (i=0;i<256;i++) {
      if (data[i] != mallocAreaX[i]) {
	  if (b == 0) {
	      fputs("\033[1m", stdout);
	      b = 1;
	  }
	  puthex(data[i]);
      } else {
	  if (b) {
	      fputs("\033[0m", stdout);
	      b = 0;
	  }
	  puthex(data[i]);
      }
      if ((i&7) == 7)
	putchar('\n');
    }
    if (b) {
	fputs("\033[0m", stdout);
	b = 0;
    }
  }
  if (spare) {
    puts("s");
    for (i=0;i<8;i++)
      puthex(spare[i]);
    putchar('\n');
  }
  memcpy(mallocAreaX, data, 256);
#endif
}

#define NandSelect() USEX(GPIO0_ODATA) &= ~GPIO0_CS1
#define NandDeselect() USEX(GPIO0_ODATA) |= GPIO0_CS1
void NandPutCommand (register __a0 u_int16 command);
void NandPutBlockAddress(register __c u_int32 addr);
void NandPutDataAddress(register __c u_int32 addr);
void NandPutLargePageSpareAddress(register __c u_int32 addr);
u_int16 NandGetStatus(void);
void NandPutAddressOctet(register __a0 u_int16 address);
void NandGetOctets(register __c0 s_int16 length, register __i2 u_int16 *buf);

u_int16 NandGetIdent(void){
  u_int16 ident[4];
  NandDeselect();
  NandSelect();
  NandPutCommand(NAND_OP_READ_SIGNATURE);
  NandPutAddressOctet(0);
  NandGetOctets(8,ident);
  NandDeselect();
  fputs("Nand Signature: ",stdout);
  /*
    2cdc 9095 5400 MT29F4G08AAA
    ecd3 14a5 64ec K9G8G08U0A-PCB0   8Gb MLC  type=3 bl=9 sz=21
    ecd5 55a5 68ec K9LAG08U0A-PCB0  16Gb MLC  type=3 bl=9 sz=22 (2 dies)
    add3 14a5 34ad HY27UT088G2A-TPC8 8Gb MLC  type=3 bl=9 sz=21
    add3 14a5 64ad HY27UU08AG5M          MLC  type=3 bl=9 sz=22(21) (2 /CS, 2 R/B)
  */
  /*
    1st byte: maker code
    2nd byte: device code
    3rd & 4th byte: parameters
   */
#if 1
  {
    register int i;
    for (i=0; i<3; i++){
      puthex(ident[i]);
    }
  }
  fputs("\n",stdout);
#endif

#if 0
  puthex(1 << ((ident[1]>>8) & 3));
  puts("dies");
  puthex(((ident[1] >> 10) & 3) + 1);
  puts("bits/cell");
  puthex(1 << ((ident[1] >> 12) & 3));
  puts("sim.prog.pages");
//  puts((ident[1] & 0x4000U) ? "Interleaved program" : "No interleaved program");
//  puts((ident[1] & 0x8000U) ? "Cache program" : "No cache program");
  puthex((ident[1] & 3)+1);
  puts("kB page");
  puthex(64 << ((ident[1] >> 4) & 3));
  puts("kB block");
#endif
  return ident[0];
}

void main(void) {
    //ph = FsPhNandCreate(1);

  while (1) {
    static char in[200];
    char *tmp;
    int mode;

    puthex(((struct FsNandPhys *)ph)->nandType);
    puts("=nT");
    puthex6(ph->eraseBlockSize);
    puts("=eraseBlockSize");
    puthex6(ph->eraseBlocks);
    puts("=eraseBlocks");
    NandGetIdent();

    puts("\n  1) Badblocks");
#ifdef DUMP_PHYS
    puts("  2) Dump phys");
#endif
#ifdef DUMP_LARGE_PAGE
    puts("  2) Dump phys (large-page)");
#endif
#ifdef DUMP_FAT
    puts("  3) Dump fat");
#endif
#ifdef DUMP_DIR
    puts("  4) Dump directory");
#endif
    //puts("100) Determine flash type");
    //puts("101) Determine nand size");
    puts("111) Erase all");
    puts("120) Init flash");
#ifdef MAP_FMT
    puts(" 50) Map fmt");
#endif
    puts(" 10) Pr sectors");
    puts(" 11) Er block");
#ifdef WR_SECTOR
    puts(" 12) Wr block");
#endif
#ifdef MARK_BAD
    puts(" 13) Mark bad");
#endif
#ifdef WR_IMAGE
    puts(" 14) program img");
#endif
#ifdef WR_BITS
    puts(" 15) write 0-bits");
#endif
    fputs("\ncommand: ", stdout);
    fflush(stdout);

    if (!fgets(in, 200, stdin))
	return;
    mode = strtol(in, &tmp, 0);
    if (tmp && *tmp && *tmp != '\n' && *tmp != '\r') {
//      puts("Inv");
      continue;
    }
    switch (mode) {
#ifdef WR_BITS
    case 15:
    {
	int i;
	u_int32 sector = 0;

	fputs("sector: ", stdout);
	fflush(stdout);
	fgets(in, 200, stdin);
	sector = strtol(in, &tmp, 0);
	if (!in[0] || (tmp && *tmp && *tmp != '\n' && *tmp != '\r'))
	  break;

	fputs("bit: ", stdout);
	fflush(stdout);
	fgets(in, 200, stdin);
	i = strtol(in, &tmp, 0);
	if (!in[0] || (tmp && *tmp && *tmp != '\n' && *tmp != '\r'))
	  break;

#if 0
	memset(minifatBuffer, -1, 512/2);
	minifatBuffer[i/16] ^= 1 << (i&15);
	ph->Write(ph, sector, 1, minifatBuffer, NULL);
#else
	memset(tmpBuf, -1, 512/2);
	tmpBuf[i/16] ^= 1 << (i&15);
	ph->Write(ph, sector, 1, NULL, tmpBuf);
#endif
	puthex(i);
	puts("Written");
    }
	break;
#endif
#ifdef WR_IMAGE
    case 14:
    {
	FILE *fp;
	int i;

	fputs("Filename: ", stdout);
	fflush(stdout);
	fgets(in, 200, stdin);
	i = strlen(in);
	while (i && (in[i-1] == '\n' || in[i-1] == '\r'))
	    in[--i] = '\0';

	fp = fopen(in, "rb");
	if (fp) {
	    u_int32 sector = 0;
	    ph->Erase(ph, 0);
	    PERIP(INT_ENABLEL) &= ~INTF_RX;
	    while (fread(minifatBuffer, 1, 512/2, fp)) {
		memset(tmpBuf, -1, 16/2);
		if (ph->Write(ph, sector, 1, minifatBuffer, tmpBuf) != 1) {
		    puts("WrError");
		}
		sector++;
		memset(minifatBuffer, -1, 512/2);
	    }
	    fclose(fp);
	    puthex(sector);
	    puts("=sectors written");
	    PERIP(INT_ENABLEL) |= INTF_RX;
	}
        break;
    }
#endif

#ifdef MARK_BAD
    case 13:
      while (1) {
	  long bad;
	  fputs("Mark eraseblock bad: ", stdout);
	  fflush(stdout);
	  fgets(in, 40, stdin);
	  bad = strtol(in, &tmp, 0);
	  if (tmp && *tmp && *tmp != '\n' && *tmp != '\r') {
	      puts("Inv");
	      break;
	  }
	  if (bad >= 1 && bad < ph->eraseBlocks) {
	      memset(minifatBuffer, -1, 16/2);
	      minifatBuffer[0+2] = 0xff00;
	      ph->Write(ph, ph->eraseBlockSize*(u_int32)bad, 1, NULL, minifatBuffer);
	  }
      }
      break;
#endif
    case 100:
	((struct FsNandPhys *)ph)->nandType =
	    (((struct FsNandPhys *)ph)->nandType + 1) & 7;
      break;

    case 10:
    {
      u_int32 sector;
      int i;

      while (1) {
	fputs("sector: ", stdout);
	fflush(stdout);
	fgets(in, 200, stdin);
	i = strlen(in);
	while (i && (in[i-1] == '\n' || in[i-1] == '\r'))
	  in[--i] = '\0';
	sector = strtol(in, &tmp, 0);
	if (!in[0] || (tmp && *tmp && *tmp != '\n' && *tmp != '\r'))
	  break;

	if (ph->Read(ph, sector, 1, minifatBuffer, tmpBuf) != 1) {
	  puthex(fsNandPhys.waitns);
	  puts("Error");
	}
	PrintSector(minifatBuffer, tmpBuf);
	puts("");
      }
      break;
    }
#ifdef WR_SECTOR
    case 12:
    {
      u_int32 sector;
      int i;

      while (1) {
	fputs("sector: ", stdout);
	fflush(stdout);
	fgets(in, 200, stdin);
	i = strlen(in);
	while (i && (in[i-1] == '\n' || in[i-1] == '\r'))
	  in[--i] = '\0';
	sector = strtol(in, &tmp, 0);
	if (!in[0] || (tmp && *tmp && *tmp != '\n' && *tmp != '\r'))
	  break;

	memset(tmpBuf, -1, 8);
	memset(minifatBuffer, -1, 256);
	minifatBuffer[0] = sector>>16;
	minifatBuffer[1] = sector;
	if (ph->Write(ph, sector, 1, minifatBuffer, tmpBuf) != 1) {
	  puts("WrError");
	}
	if (ph->Read(ph, sector, 1, minifatBuffer, tmpBuf) != 1) {
	  puts("RdError");
	}
	PrintSector(minifatBuffer, tmpBuf);
	puts("");
      }
      break;
    }
#endif
    case 11:
    {
      u_int32 sector;
      int i;

      while (1) {
	fputs("Erase sector: ", stdout);
	fflush(stdout);
	fgets(in, 200, stdin);
	i = strlen(in);
	while (i && (in[i-1] == '\n' || in[i-1] == '\r'))
	  in[--i] = '\0';
	sector = strtol(in, &tmp, 0);
	if (!in[0] || (tmp && *tmp && *tmp != '\n' && *tmp != '\r'))
	  break;

	if (ph->Erase(ph, sector)) {
	  puts("Error");
	}
	//PrintSector(minifatBuffer, tmpBuf);
	//puts("");
      }
      break;
    }
    case 111:
      /* Erase all */
    {
      u_int32 sector, sectors = (u_int32)ph->eraseBlockSize * ph->eraseBlocks;
      int i;
      FILE *fp;

      puthex6(ph->eraseBlockSize);
      puts("=eraseBlockSize");
      puthex6(ph->eraseBlocks);
      puts("=eraseBlocks");
      puthex6(sectors);
      puts("=sectors");

      for (sector = 0x0; sector < sectors; sector += ph->eraseBlockSize) {
	if (ph->Erase(ph, sector)) {
	  puthex6(sector);
	  puts("Error");
#if 1
	  memset(tmpBuf, 0xffffU, 8);
	  tmpBuf[0+2] = 0;
	  ph->Write(ph, sector, 1, NULL, tmpBuf);
#endif
	}
      }
      break;
    }

    case 51:
	if (!(ph->eraseBlocks & 1))
	    ph->eraseBlocks--;
	/* fall through */
#ifdef MAP_FMT
    case 50:
#if 1
	puthex(ph->eraseBlockSize);
	putchar(' ');
	puthex(ph->eraseBlocks);
	puts("=block size,erase blocks");
      map = FsMapFlCreate(ph, 1); /* format */
      map->Flush(map, 1);
      map->Delete(map);
      puthex(map->blocks>>16);
      puthex(map->blocks);
      puts("=map blocks: Mapper format complete");
#endif
      break;
#endif

    case 1:
      /* Scan for bad blocks */
    {
      u_int32 sector, sectors = (u_int32)ph->eraseBlockSize * ph->eraseBlocks;
      int i;
#if 0
      FILE *fp;
      fputs("Filename (or empty): ", stdout);
      fflush(stdout);
      fgets(in, 200, stdin);
      i = strlen(in);
      while (i && (in[i-1] == '\n' || in[i-1] == '\r'))
	in[--i] = '\0';
#endif
      puthex6(ph->eraseBlockSize);
      puts("=eraseBlockSize");
      puthex6(ph->eraseBlocks);
      puts("=eraseBlocks");
      puthex6(sectors);
      puts("=sectors");
#if 0
      if (in[0] && (fp = fopen(in, "wb"))) {
	fputs("Dumping bad blocks to \"", stdout);
	fputs(in, stdout);
	puts("\"\n");
      } else {
	fp = stdout;
      }
#endif
      puts("Scanning Bad Blocks");
      for (sector = 0x0; sector < sectors;
#if 0
	   sector += 4
#else
	       sector += ph->eraseBlockSize
#endif
	  ) {
	if (ph->Read(ph, sector, 4, NULL, tmpBuf) != 4) {
	}
	if ((tmpBuf[0+2] & 0xff) != 0xff
	    || (tmpBuf[8+2] & 0xff) != 0xff
	    || (tmpBuf[16+2] & 0xff) != 0xff
	    || (tmpBuf[24+2] & 0xff) != 0xff
	    ) {
	    fputhex6(sector, stdout/*fp*/);
	    fputs("=B\n", stdout/*fp*/);
	}
      }
#if 0
      if (fp && fp != stdout)
	fclose(fp);
#endif
      puts("Done\n");
      break;
    }
#ifdef DUMP_LARGE_PAGE
    case 2:
      /* dump physical - large page mode */
    {
      u_int32 sector, sectors =
	(u_int32)ph->eraseBlockSize * ph->eraseBlocks;
      u_int32 __y na = 0;
      FILE *fp;
      int i;

      fputs("Filename: ", stdout);
      fflush(stdout);
      fgets(in, 200, stdin);
      i = strlen(in);
      while (i && (in[i-1] == '\n' || in[i-1] == '\r'))
	in[--i] = '\0';
#if 0
      puthex6(ph->eraseBlockSize);
      puts("=eraseBlockSize");
      puthex6(ph->eraseBlocks);
      puts("=eraseBlocks");
#endif
      puthex6(sectors);
      puts("=sectors");

      fp = fopen(in, "wb");
      if (fp) {
	fputs("Dumping Physical (large-page) to \"", stdout);
	fputs(in, stdout);
	puts("\"\n");
	for (sector = 0x0; sector < sectors; sector += 4) {
	  if ((sector & 0x1ff) == 0) {
	    putchar('\r');
	    puthex6(sector);
	    putchar(' ');
	    puthex6(na);
	    fflush(stdout);
	  }
	  /*Note: ph->Read(ph, sector, 4, mallocAreaX, tmpBuf);
	          does not work! */
	  ph->Read(ph, sector, 4, NULL, tmpBuf);
	  if (ph->Read(ph, sector, 4, mallocAreaX, NULL) != 4) {
	    na++;
#if 0
	    puthex6(sector);
	    puts(" N/A");
#endif
	  }
	  fwrite(mallocAreaX, 4*512/2, 1, fp);
	  /* Swap all back to make it valid for NAND programmer */
	  NandSwapBad(tmpBuf+0);
	  NandSwapBad(tmpBuf+8);
	  NandSwapBad(tmpBuf+16);
	  NandSwapBad(tmpBuf+24);
	  fwrite(tmpBuf, 4*16/2, 1, fp);
	}
	fclose(fp);
	putchar('\n');
	puthex6(na);
	puts("=N/A\nDump done.");
      } else {
#if 0
	fputs("Could not open \"", stdout);
	fputs(in, stdout);
	puts("\" for writing\n");
#endif
      }
      break;
    }
#endif
#ifdef DUMP_PHYS
    case 2:
      /* dump physical */
    {
      u_int32 sector, sectors =
	(u_int32)ph->eraseBlockSize * ph->eraseBlocks;
      u_int32 __y na = 0;
      FILE *fp;
      int i;

      fputs("Filename: ", stdout);
      fflush(stdout);
      fgets(in, 200, stdin);
      i = strlen(in);
      while (i && (in[i-1] == '\n' || in[i-1] == '\r'))
	in[--i] = '\0';

      puthex6(ph->eraseBlockSize);
      puts("=eraseBlockSize");
      puthex6(ph->eraseBlocks);
      puts("=eraseBlocks");
      puthex6(sectors);
      puts("=sectors");

      fp = fopen(in, "wb");
      if (fp) {
#if 0
	fputs("Dumping Physical to \"", stdout);
	fputs(in, stdout);
	puts("\"\n");
#endif
	for (sector = 0x0; sector < sectors; sector++) {
	  if ((sector & 0x1ff) == 0) {
	    putchar('\r');
	    puthex6(sector);
	    //putchar(' ');
	    //puthex6(na);
	    fflush(stdout);
	  }

#if 0
	  ph->Read(ph, sector, 1, NULL, tmpBuf);
	  ph->Read(ph, sector, 1, minifatBuffer, NULL);
	  puthex(PERIP(NFLSH_LPL));
	  puthex(PERIP(NFLSH_CP_LPH));
	  puts("=LPL,CP,LPH");
	  if (0)
#else
	  if (ph->Read(ph, sector, 1, minifatBuffer, tmpBuf) != 1)
#endif
	  {
	    na++;
#if 0
	    puthex6(sector);
	    puts(" N/A");
#endif
	  }
//	  fwrite(tmpBuf, 16/2, 1, fp);
	  fwrite(minifatBuffer, 512/2, 1, fp);
	}
	fclose(fp);
	putchar('\n');
	//puthex6(na);
	puts("=N/A\nDump done.");
      } else {
#if 0
	fputs("Could not open \"", stdout);
	fputs(in, stdout);
	puts("\" for writing\n");
#endif
      }
      break;
    }
#endif

#ifdef DUMP_FAT
    case 3:
    {
      u_int32 sector, sectors;
      FILE *fp;
      int i;

      fputs("Filename: ", stdout);
      fflush(stdout);
      fgets(in, 200, stdin);
      i = strlen(in);
      while (i && (in[i-1] == '\n' || in[i-1] == '\r'))
	in[--i] = '\0';

      puts("Tiny-Mapper");
      map = FsMapTnCreate(ph, 0); /* tiny mapper */
      sectors = map->blocks;

      /* Dump fat */
      puthex6(sectors);
      puts("=sectors");

      fp = fopen(in, "wb");
      if (fp) {
	fputs("Dumping Logical (FAT) to \"", stdout);
	fputs(in, stdout);
	puts("\"\n");
	for (sector = 0x0; sector < sectors; sector++) {
	  if ((sector & 0x1ff) == 0) {
	    putchar('\r');
	    puthex(sector);
	    fflush(stdout);
	  }
	  if (map->Read(map, sector, 1, minifatBuffer) != 1) {
	    memset(minifatBuffer, 0, 256);
#if 0
	    puthex6(sector);
	    puts(" N/A");
#endif
	  }
	  fwrite(minifatBuffer, 512/2, 1, fp);
	}
	fclose(fp);
	puts("Dump done");
      } else {
	fputs("Could not open \"", stdout);
	fputs(in, stdout);
	puts("\" for writing\n");
      }
      map->Delete(map);
      break;
    }
#endif

#ifdef DUMP_DIR
    case 4:
    {
      u_int32 sector, sectors;
      int i;

      puts("Tiny-Mapper");
      map = FsMapTnCreate(ph, 0); /* tiny mapper */

      if (InitFileSystem() == 0) {
	  player.totalFiles = OpenFile(0xffffU);
	  puthex(player.totalFiles);
	  puts("=totalFiles");
	  for (i=0;i<player.totalFiles;i++) {
	      int j;
	      OpenFile(i);
	      puthex6(minifatInfo.fileSize);
	      putchar(' ');
	      for (j=0;j<6;j++) {
		  putchar(minifatInfo.fileName[j]>>8);
		  putchar(minifatInfo.fileName[j]);
	      }
	      puts("");
	  }
      } else {
	puts("InitFileSystem() failed for writing\n");
      }
      break;
    }
#endif

    case 120:
    {
      int blockSizeExp = 5, flashSizeExp = 18;
      long wsns = GetParam("Waits (ns): ");
      __y long tmp;
      if (wsns == NAN)
	break;

      tmp = GetParam("NandType: ");
      if (tmp == NAN)
	break;
      ((struct FsNandPhys *)ph)->nandType = tmp;
      ((struct FsNandPhys *)ph)->waitns = wsns;

      tmp = GetParam("blockSizeExpr: ");
      if (tmp == NAN)
	break;
      blockSizeExp = tmp;
      ph->eraseBlockSize = (1 << blockSizeExp);
      tmp = GetParam("flashSizeExpr: ");
      if (tmp == NAN)
	break;
      flashSizeExp = tmp;
      ph->eraseBlocks = 1 << (flashSizeExp-blockSizeExp);

      puthex(1<<blockSizeExp);
      puts("=eraseBlockSize");
      puthex(ph->eraseBlocks);
      puts("=eraseBlocks");


      ph->Erase(ph, 0);
      memset(minifatBuffer, -1/*0xffffU*/, 256);
      minifatBuffer[0] = ('V' << 8) + 'L';
      minifatBuffer[1] = ('S' << 8) + 'I';
      minifatBuffer[2] = ((struct FsNandPhys *)ph)->nandType;
      minifatBuffer[3] = (blockSizeExp << 8) + flashSizeExp;
      minifatBuffer[4] = ((struct FsNandPhys *)ph)->waitns;
      memset(tmpBuf, -1, 16);

      ph->Write(ph, 0, 1, minifatBuffer, tmpBuf);
      puts("Inited");
      break;
    }

    default:
      puthex(mode);
      puts("mode");
    }
  }
}

/*
  Hynix HY27UT084G2 MLC 512MB

0100  0fff =block size,erase blocks
000e 6500 =map blocks: Mapper format complete
0003 =nT
000100=eraseBlockSize
000fff=eraseBlocks


block 256:
Data  ...
Spare 0000 0001 ffff ffff 0100 0000 ffff ffff 
      -------ECC     ----unused
             --type=ROOT  ---------logical page number 0x00000100 = 256
	        --reserved          ---------new branch (page# for root node)
		  --bad block


FlashSize 1G: bits 21
Data blocks: upto 2^21 0x4182..0xffeff first and last erase block reserved
Branch4: upto 2^14  0x182..0x4181
Branch3: upto 2^7   0x102..0x181
Branch2: upto 2^0   0x101
Root:    1          0x100

 */
